
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <str.h>

struct string {
	usize len, cap;
	struct str_meta {	
		enum str_ownership owned;
		union _str_meta_inner {
			// `STRF_OWNED` 
			struct str_meta_owned {
				_Atomic usize refs; // Number of derrivations from this 
				void (*m_free)(void*); // How to free (when `_MALLOC`).
				_Atomic bool orphan; // If a `derive` references this parent somehow, but this string has been `str_free()`d. Tell the derived string that sets the `refs` to 0 to free it.
			} owned;
			// `STRF_DERIVED`
			struct str_meta_derived {
				// String this is derived from, can be either owned or derived.
				struct string* _shared parent;
				// `str`'s offset from `parent->str`.
				usize offset;
			} derived;
		} _inner;
	} meta;
	// The start of this string slice. Regardless of slicing
	char* str;
};

//TODO: We can use offsetof() to reverse the result of `str_new()` from char* (str field) to str_t*.



inline static bool _str_meta_is_owned(enum str_ownership owned)
{
	return AS(owned, u32) & AS(STRF_OWNED, u32);
}

// Returns true for owned, false for derived
inline static  bool _str_meta_parts_from(enum str_ownership o, union _str_meta_inner *pIN inner, struct str_meta_owned* *pOUT own, struct str_meta_derived* *pOUT derive)
{
	if(_str_meta_is_owned(o))
		return (PTR_ASSIGN(own, &inner->owned), true);
	else	return (PTR_ASSIGN(derive, &inner->derived), false);
}

inline static bool _str_is_owned(const str_t*pIN str)
{
	return _str_meta_is_owned(str->meta.owned);
}

inline static bool _str_is_owned_parts(str_t*pIN str, struct str_meta_owned* *pOUT own, struct str_meta_derived* *pOUT derive)
{
	return _str_meta_parts_from(str->meta.owned, &str->meta._inner, own, derive);
}

static union _str_meta_inner _str_default_meta(enum str_ownership oship, bool *pOUT owned)
{
	union _str_meta_inner output;
	
	struct str_meta_owned* o_own = NULL;
	struct str_meta_derived* o_der = NULL;

	// PTR_ASSIGNx always evaluates both args, so this is fine
	PTR_ASSIGNx(owned,
			_str_meta_parts_from(oship, &output, 
				&o_own,
				&o_der)
	);
	switch(oship)
	{
		case STR_OWNED_STACK:
		case STR_OWNED_STATIC:
			debug_assert(o_own);
			*o_own = (struct str_meta_owned){ 
				.refs = 0,
				.m_free = NULL,	
				.orphan = false,
			};
			break;
		case STR_OWNED_MALLOC:
			debug_assert(o_own);
			*o_own = (struct str_meta_owned){
				.refs = 0,
				.m_free = &free,
				.orphan = false,
			};
			break;

		case STR_DERIVED_STATIC:
		case STR_DERIVED_STACK:
		case STR_DERIVED_MALLOC:
			debug_assert(o_der);
			*o_der = (struct str_meta_derived){
				.parent = NULL,
				.offset = 0,
			};
			break;
		default:
			ERROR("Cannot assign metadata from ownership type 0x%x, returning 0'd metadata", AS(oship, u32));	
			memset(&output, 0, sizeof(output));
	}
	return output;
}

/// Allocates a bare new str_t. The metadata will be incomplete for `derived` string slices, and `str` will be `NULL`. For `owned` strings, `str` will be `calloc()`ed to `cap+1`.
static str_t _str_alloc_bare(usize cap, enum str_ownership owned)
{
	bool is_owned;
	let meta_inner = _str_default_meta(owned, &is_owned);
	struct str_meta meta = {
		.owned = owned,
		._inner = meta_inner,
	};
	str_t str = {
		.len = 0,
		.cap = cap,
		.meta = meta,
	};
	str.str = is_owned ? calloc(cap+1, 1) : NULL;
	return str;
}

static void _str_force_free(struct str_meta_owned*pIN o_own, void* ptr)
{
	(o_own->m_free ?: &free)(ptr);
}

void str_free(str_t* restrict strp)
{
	str_t str = unbox_value(strp);
	
	struct str_meta_owned* o_own = NULL;
	struct str_meta_derived* o_der = NULL;

	if(_str_is_owned_parts(&str, &o_own, &o_der))
	{
		debug_assert(o_own);
		if(o_own->refs > 0) o_own->orphan = true;
		else if(str.meta.owned == STR_OWNED_MALLOC && str.str)
			_str_force_free(o_own, str.str);
	}
	else {
		debug_assert(o_der);
		TODO(" \
		//TODO: Find top-level parent (traverse list of potentially derived `o_der->parent`s. \
		//TODO: While doing so, compute the full offset that this slice has from the top owned parent. \
		//TODO: Then, decrement its `refs` \
		//TODO: Then, if it is marked as `orphan`, and we have decremented its `refs` to 0, free its `ptr`. \
		");
	}
}
